/*  $Id: cache.c,v 1.1 1997/12/05 07:35:49 news Exp news $
**
**  History cache routines.
*/
#include <stdio.h>
#include <ctype.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/param.h>
#include <syslog.h>
#include <limits.h>
#include <time.h>
#include "configdata.h"
#include "qio.h"
#include "cache.h"

#ifdef	DO_DEBUG
#define	DEBUG(args)	{(void) printf args;}
#else
#define	DEBUG(args)	;
#endif

#ifdef	CACHE_MAIN
#define	TELL_NOW(fmt, now)	(void)fprintf(stderr, fmt, now)
#else
#define	TELL_NOW(fmt, now)	syslog(LOG_NOTICE, fmt, now)
#endif

/*
  This file provides a series of subroutines for caching message ids.
  The cache differs from the conventional one in the way that it keeps
  more recent ids than the older.

  A message entry consists of three fields, the hash, the existence flag,
  and the arrival time.  The time field is what the conventional history
  cache lacks.

  If we use 15-bit wide time word and the unit is in seconds, the max
  time will be 32767 second, i.e., 9.2 hours.  The number of bits
  of the time word is defined as CCH_TIMEBITS.  Furthermore, there's
  no need to maintain accurate times, i.e., we can drop some of the least
  significat time bits.  We give the name CCH_TIMESHIFT to this number
  of dropped bits.

  The cache time field stores the relative time to the cache epoch.
  However, as time goes by, the time field will eventually overflow.
  We have to introduce the variable "cacheBaseTime", and it must be
  incresed during the cache expire.  We expire the cache if the
  time value overflows.  The new caches will use larger base time.

  The cache base time is assigned during the cache_init and cache_expire.
  However, if it is too small, there can be too frequent time overflows,
  resulting too frequent cache_expire.  It is the most expensive
  process among our cache routine.  On the contrary, if the base time
  is too close to NOW, previous times will be lost too soon.
  We choose the delta(CCH_BASE_DELTA) as one half of the CCH_TIME_MAX.

  Some of the important variables are as follows:

  cacheTimeUnit  : cache time unit in seconds [8]
		   all internal times have this time unit.

  cacheBaseTime  : the base time of this cache.

*/
 
/********** Start of configurable constants *******************************/

#define CCH_TIMEBITS		15
#define CCH_TIMESHIFT		3
#define CCH_PROBEMAX		100

/*** contents of cache.h ************************************************

#define HIS_CACHE_SIZE		(256 * 1024)

typedef enum {CCH_HAVEART, CCH_HAVEID, CCH_NO} cachelook_t;

typedef struct {
    unsigned long size;
    BOOL     initialized;
    unsigned long positive;
    unsigned long negative;
    unsigned long oldest;
    unsigned long newest;
} cachestat_t;

void CCHinit(CSTRING history);
void CCHstore(CSTRING msgid, time_t posted, BOOL flag);
cachelook_t CCHhavearticle(CSTRING msgid);
U_64 CCHhash(const void *str);
BOOL CCHsync(CSTRING history);

*** end of cache.h *****************************************************/

#define CCH_TIMEMASK		( (1 << CCH_TIMEBITS) - 1 )
#define	CCH_BASE_DELTA	( 1 << (CCH_TIMEBITS - 1) )
#define CCH_FLAGMASK		( 1 << CCH_TIMEBITS )
#define CCH_HASHMASK		( ~(U_64)(CCH_FLAGMASK | CCH_TIMEMASK) )

#define CCH_TOPOS(x)	  ( (unsigned long)((x)&ULONG_MAX) % HIS_CACHE_SIZE )
#define CCH_TOHASH(x)		( (x) & CCH_HASHMASK )
#define CCH_TOTIME(x)		( (time_t)((x) & ULONG_MAX) & CCH_TIMEMASK )
#define CCH_TOFLAG(x)		( (unsigned long)((x) & (U_64)CCH_FLAGMASK) )
#define CCH_TOHAFL(x)		( (x) & ~(U_64)CCH_TIMEMASK )

#define NOTFOUND		-1
#define CCH_TIMEMAX		CCH_TIMEMASK

/*
 * internal routines
 */

static BOOL cacheInitialized = FALSE;
static time_t cacheBaseTime;
static U_64 cachetable[HIS_CACHE_SIZE];
typedef struct {
    int size;		/* table size */
    int timebits;
    int timeshift;
    int probemax;
    BOOL initialized;
    time_t basetime;
} cacheHeader;
static cacheHeader ch;
  
/*
 * This is a simplified version of the pathalias hashing function.
 * Thanks to Steve Belovin and Peter Honeyman
 *
 * hash a string into a long int.  31 bit crc (from andrew appel).
 * the crc table is computed at run time by crcinit() -- we could
 * precompute, but it takes 1 clock tick on a 750.
 *
 * This fast table calculation works only if POLY is a prime polynomial
 * in the field of integers modulo 2.  Since the coefficients of a
 * 32-bit polynomial won't fit in a 32-bit word, the high-order bit is
 * implicit.  IT MUST ALSO BE THE CASE that the coefficients of orders
 * 31 down to 25 are zero.  Happily, we have candidates, from
 * E. J.  Watson, "Primitive Polynomials (Mod 2)", Math. Comp. 16 (1962):
 *	x^32 + x^7 + x^5 + x^3 + x^2 + x^1 + x^0
 *	x^31 + x^3 + x^0
 *
 * We reverse the bits to get:
 *	111101010000000000000000000000001 but drop the last 1
 *         f   5   0   0   0   0   0   0
 *	010010000000000000000000000000001 ditto, for 31-bit crc
 *	   4   8   0   0   0   0   0   0
 */

/*
** 31-bit polynomial (avoids sign problems)
** #define POLY 0x48000000L
*/
#define POLY 0xf5000000L	/* 32-bit polynomial */

static unsigned long CrcTable[128];

/*
 - crcinit - initialize tables for hash function
 */
static void
crcinit()
{
    int i, j;
    unsigned long sum;

    for (i = 0; i < 128; ++i) {
	sum = 0L;
	for (j = 7 - 1; j >= 0; --j)
	    if (i & (1 << j))
		sum ^= POLY >> j;
	CrcTable[i] = sum;
    }
    DEBUG(("crcinit: done\n"));
}

static U_64
cache_hash_id(name_in, size_in)
char *name_in;
int size_in;
{
    char *name = name_in;
    int size;
    unsigned long sum = 0L;
    unsigned long high;

    size = size_in;
    while (size--) {
	sum = (sum >> 7) ^ CrcTable[(sum ^ (*name++)) & 0x7f];
    }
    /*
     * assign first hash.  The bit 31 is always set.
     */
    high = sum | (1 << 31);

    size = size_in;
    while (size--) {
	sum = (sum >> 7) ^ CrcTable[(sum ^ (*--name)) & 0x7f];
    }

    /*
     * Pack two 32 bit hashes.
     */
    return ((U_64)high << 32) | (U_64)sum;
}

static char *cipoint(char *s, size_t size)
{
    char *p;
    static char post[] = "postmaster";
    static int plen = sizeof(post) - 1;

    if ((p = memchr(s, '@', size)) == NULL)
	return NULL;	/* assume all local */
    if ((p - (s + 1) == plen) && !strncasecmp(post, s+1, plen)) {
	/* crazy -- "postmaster" is case-insensitive */
	return s;
    }
    return p;
}

U_64
CCHhash(const char *msg)
{
    int		len;
    char	*p, *str, buf[256];
    U_64	hash;
    void	free();

    len = strlen(msg);
    str = (len < sizeof buf) ? strcpy(buf, msg) : strdup(msg);
    if ((p = cipoint(str, len)))
	while(*++p)
	    *p = tolower(*p);
    hash = cache_hash_id(str, len);
    if (str != buf)
	free(str);
    return hash;
}

static time_t time_to_cachetime(time_t t)
{
    t >>= CCH_TIMESHIFT;
    return (t > cacheBaseTime) ? (t - cacheBaseTime) : 0;
}

static time_t cachetime_to_time(time_t ct)
{
    return (ct + cacheBaseTime) << CCH_TIMESHIFT;
}

static void cache_expire(time_t now)
{
    time_t newbase;
    time_t dt;
    time_t then;
    int i;

    TELL_NOW("SERVER expireing cache\n", now);
    then = now >>= CCH_TIMESHIFT;
    newbase = then - CCH_BASE_DELTA;

    if (then < CCH_BASE_DELTA || newbase < cacheBaseTime) {
	TELL_NOW("SERVER cache expire time %lu too small\n", now);
	return;
    }

    dt = newbase - cacheBaseTime;
    for (i=0; i<HIS_CACHE_SIZE; i++) {
	then = CCH_TOTIME(cachetable[i]);
	if (then > dt)
	    cachetable[i] = CCH_TOHAFL(cachetable[i]) | (U_64)(then - dt);
	else
	    cachetable[i] = CCH_TOHAFL(cachetable[i]);
    }
    cacheBaseTime = newbase;
}

/**********************************************************************
**
** Start of public routines
**
**********************************************************************/

/*
 * Do I have the id in cache?
 */
cachelook_t
CCHhavearticle(char *s)
{
    U_64		hash;
    unsigned long	offset;
    int			i;

    hash   = CCHhash(s);
    offset = CCH_TOPOS(hash);
    hash   = CCH_TOHASH(hash);

    for (i=0; i<CCH_PROBEMAX; i++) {
	if (hash == CCH_TOHASH(cachetable[offset]))
	    return (CCH_TOFLAG(cachetable[offset]) != 0) ?
		CCH_HAVEART : CCH_HAVEID;
	if (++offset == HIS_CACHE_SIZE)
	    offset = 0;
    }
    return CCH_NO;
}

/*
** Write an entry to the cache.
*/

void
CCHstore(CSTRING s, time_t posted, BOOL flag)
{
    U_64		rawhash;		/* raw hash */
    U_64		hash_h;			/* masked hash */
    unsigned long	offset;
    int			i;
    time_t		now;
    time_t		then;
    time_t		oldest;
    unsigned long	oldestoffset;

    rawhash = CCHhash(s);
    hash_h = CCH_TOHASH(rawhash);
    offset = CCH_TOPOS(rawhash);

    oldestoffset = offset;
    for (i=0; i<CCH_PROBEMAX; i++) {

	if (hash_h == CCH_TOHASH(cachetable[offset]))
	    break;
	
	then = CCH_TOTIME(cachetable[offset]);

	if (i == 0)
	    oldest = then;
	else if (then < oldest)	{ /* remember the slot for maximum dt. */
	    oldest = then;
	    oldestoffset = offset;
	}

	/* now next prepare for the next slot */
	if (++offset >= HIS_CACHE_SIZE)
	    offset = 0;
    }

    if (i == CCH_PROBEMAX) {
	offset = oldestoffset;
    }

    DEBUG(("oldcache: %s\n", cache_to_str(cachetable[offset])));

    while ((now = time_to_cachetime(posted)) > CCH_TIMEMAX)
	cache_expire(posted);

    if (flag)
	cachetable[offset] = hash_h | (U_64)(CCH_FLAGMASK | now);
    else
	cachetable[offset] = hash_h | (U_64)now;

    DEBUG(("newcache: %s\n", cache_to_str(cachetable[offset])));
}

/*
** read cache data from the file
*/
BOOL
cache_readfile(CSTRING history)
{
    char path[512];
    FILE *fpi;
    int i, j;

    sprintf(path, "%s.cch", history);
    if ((fpi = fopen(path, "r")) == NULL)
	return FALSE;

    i = fread(&ch, sizeof(ch), 1, fpi);
    if (i == 1 &&
	        ch.size == HIS_CACHE_SIZE &&
		ch.timebits == CCH_TIMEBITS &&
		ch.timeshift == CCH_TIMESHIFT &&
		ch.probemax == CCH_PROBEMAX) {
	j = fread(cachetable, sizeof(cachetable), 1, fpi);
    }
    fclose(fpi);
    cacheInitialized = ch.initialized;
    cacheBaseTime = ch.basetime;
    return (i == 1 && j == 1);
}

void
CCHinit(CSTRING history)
{
    time_t now;

    if (cacheInitialized)
	return;

    crcinit();

    if (cache_readfile(history)) {
	TELL_NOW("SERVER cache reloaded with basetime %lu\n",
		(cacheBaseTime << CCH_TIMESHIFT));
	return;
    }

    now = time(NULL);
    now = now >> CCH_TIMESHIFT;
    cacheBaseTime = now - CCH_BASE_DELTA;
    memset(cachetable, 0, sizeof(cachetable));
    cacheInitialized = TRUE;
    TELL_NOW("SERVER cache started with empty data\n", now);
}

/*
** write current cache data to the file
*/
BOOL
CCHsync(CSTRING history)
{
    char path[512];
    FILE *fpo;
    int i;

    sprintf(path, "%s.cch", history);
    if ((fpo = fopen(path, "w")) == NULL)
	return FALSE;
     
    ch.size = HIS_CACHE_SIZE;
    ch.timebits = CCH_TIMEBITS;
    ch.timeshift = CCH_TIMESHIFT;
    ch.probemax = CCH_PROBEMAX;
    ch.initialized = cacheInitialized;
    ch.basetime = cacheBaseTime;

    i = fwrite(&ch, sizeof(ch), 1, fpo);
    if (i == 1)
	i = fwrite(cachetable, sizeof(cachetable), 1, fpo);
    fclose(fpo);
    return (i == 1);
}

#if	defined(CACHE_MAIN) && !defined(CACHE_STAT)
#define CACHE_STAT
#endif

#ifdef	CACHE_STAT

cachestat_t *CCHstat()
{
    static cachestat_t s;
    time_t then;
    int    i;

    s.size = HIS_CACHE_SIZE;
    s.initialized = cacheInitialized;
    s.oldest = CCH_TIMEMAX;
    s.newest = 0;
    s.positive = 0;
    s.negative = 0;

    for (i=0; i<HIS_CACHE_SIZE; i++) {
	if ( CCH_TOHASH(cachetable[i]) == (U_64)0 )
	    continue;
	if (CCH_TOFLAG(cachetable[i]) != 0)
	    s.positive++;
	else
	    s.negative++;
	then = CCH_TOTIME(cachetable[i]);
	if (s.oldest > then)
	    s.oldest = then;
	if (s.newest < then)
	    s.newest = then;
    }
    s.oldest = cachetime_to_time(s.oldest);
    s.newest = cachetime_to_time(s.newest);
    return &s;
}

#endif	/* CACHE_STAT */

#ifdef CACHE_MAIN

long atol();
int atoi();

/*
**  Split a line at a specified field separator into a vector and return
**  the number of fields found, or -1 on error.
*/
static int
EXPsplit(p, sep, argv, count)
    register char	*p;
    register char	sep;
    register char	**argv;
    register int	count;
{
    register int	i;

    for (i = 1, *argv++ = p; *p; )
	if (*p++ == sep) {
	    if (++i == count)
		/* Overflow. */
		return -1;
	    p[-1] = '\0';
	    for (*argv++ = p; *p == sep; p++)
		continue;
	}
    return i;
}

/*
 * Build cache from the existing history.
 */
static BOOL
build_cache(char *history, int maxline)
{
    int  i;
    int  initialized = 0;
    char line[QIO_BUFFER], *p, *fields[4];
    time_t Arrived;
    struct timeval  start_time, end_time;
    unsigned long elapsed;
    FILE *fpi;
    cachestat_t *cs;
    unsigned long unused;
    unsigned long lnum;

    /* NULL history is acceptable.  We started a new system. */
    if (history == NULL) {
	fprintf(stderr, "CCH: warning null history\n");
	return FALSE;
    }
    if ((fpi = fopen(history, "r")) == NULL) {
	fprintf(stderr, "CCH: can't open history %s\n", history);
	return FALSE;
    }

    gettimeofday(&start_time, NULL);

    for ( lnum=0; fgets(line, sizeof(line), fpi); ) {

	/* Split up the major fields. */
	i = EXPsplit(line, HIS_FIELDSEP, fields, sizeof(fields));
	if (i != 2 && i != 3) {
	    (void)fprintf(stderr, "bad line: %s\n", line);
	    continue;
	}

	/* Split up the time field, robustly. */
	if ((p = strchr(fields[1], HIS_SUBFIELDSEP)))
	    *p = '\0';
	Arrived = atol(fields[1]);
	DEBUG(("%ld %s\n", Arrived, fields[0]));

	if (!initialized) {
	    CCHinit(Arrived);
	    initialized = 1;
	}

	if (CCHhavearticle(fields[0]) != CCH_NO) {
	    fprintf(stderr, "CCHinit: skipping duplicate %s\n", fields[0]);
	} else {
	    CCHstore(fields[0], Arrived, TRUE);
	}

	if (maxline > 0 && ++lnum >= maxline)
	    break;
    }
    fclose(fpi);

    gettimeofday(&end_time, NULL);
    elapsed = (end_time.tv_sec - start_time.tv_sec) * 1000
	    + (end_time.tv_usec - start_time.tv_usec) / 1000;
    cs = CCHstat();
    unused = cs->size - (cs->positive + cs->negative);
    printf("table size %lu pos %lu neg %lu unused %lu t-range %lu\n",
	cs->size, cs->positive, cs->negative, unused, cs->newest - cs->oldest);

    lnum = cs->size - unused;
    printf("total entry %lu total time %lu msec time/entry %.3f msec\n",
	    lnum, elapsed, (double)elapsed /(double)lnum);

    return TRUE;
}

#ifdef DO_DEBUG

static void
u64_to_str(U_64 hash, char str[20])
{
    unsigned long buf[2];
    memcpy(buf, &hash, 8);
    sprintf(str, "%8.8lx %8.8lx", buf[0], buf[1]);
}

static char *cache_to_str(U_64 cache)
{
    U_64		hash;
    unsigned long	offset;
    BOOL		flag;
    char		buf1[20], buf2[20];
    static char		buff[128];

    hash = CCH_TOHASH(cache);
    offset = CCH_TOPOS(cache);
    flag = CCH_TOFLAG(cache) != 0;

    u64_to_str(cache, buf1);
    u64_to_str(hash, buf2);
    sprintf(buff, "cache %s hash %s flag %d offset %lu",
	buf1, buf2, flag, offset);
    return buff;
}

#endif	/* DO_DEBUG */

static BOOL
verify_cache(char *history, int maxline)
{
    int  i;
    unsigned long  lnum;
    char line[2048], *p, *fields[4];
    time_t Arrived;
    struct timeval  start_time, end_time;
    unsigned long elapsed, positive, negative, miss;
    FILE *fpi;
    cachelook_t cstatus;

    if ((fpi = fopen(history, "r")) == NULL) {
	fprintf(stderr, "can't open %s\n", history);
	return FALSE;
    }

    gettimeofday(&start_time, NULL);
    positive = negative = miss = 0;

    for (lnum=0; fgets(line, sizeof(line), fpi);) {

	/* Split up the major fields. */
	i = EXPsplit(line, HIS_FIELDSEP, fields, sizeof(fields));
	if (i != 2 && i != 3) {
	    (void)fprintf(stderr, "bad line: %s\n", line);
	    continue;
	}

	/* Split up the time field, robustly. */
	if ((p = strchr(fields[1], HIS_SUBFIELDSEP)))
	    *p = '\0';
	Arrived = atol(fields[1]);

	cstatus = CCHhavearticle(fields[0]);
	if (cstatus == CCH_HAVEART)
	    positive++;
	else if (cstatus == CCH_HAVEID)
	    negative++;
	else if (cstatus == CCH_NO)
	    miss++;

	if (maxline > 0 && ++lnum >= maxline)
	    break;
    }
    gettimeofday(&end_time, NULL);
    elapsed = (end_time.tv_sec - start_time.tv_sec) * 1000
	    + (end_time.tv_usec - start_time.tv_usec) / 1000;

    printf("total time %lu msec time/entry %.3f msec\n",
		elapsed, (double)elapsed /(double)lnum);

    printf("cache lookups %lu pos %lu neg %lu miss %lu\n",
		lnum, positive, negative, miss);
    fclose(fpi);
    return TRUE;
}

static void usage()
{
    fprintf(stderr, "usage: cachetest [-n num] <history>\n");
    fprintf(stderr, "  lines    process first num lines of history.\n");
    fprintf(stderr, "  history  history text file\n");
    exit(1);
}

void main(ac, av)
int  ac;
char *av[];
{
    int  i;
    char *history = NULL;
    int maxline = 0;
    extern char *optarg;
    extern int optind;

    while ((i = getopt(ac, av, "n:")) != EOF)
        switch (i) {
        default:
            usage();
            /* NOTREACHED */
        case 'n':
	    maxline = atoi(optarg);
	    break;
	}
    ac -= optind;
    if (ac != 1)
	usage();

    history = av[optind];

    printf("tablesize %d probemax %d\n", HIS_CACHE_SIZE, CCH_PROBEMAX);
    if (build_cache(history, maxline))
	(void)verify_cache(history, maxline);
}

#endif	/* CACHE_MAIN */
