/*  $Revision: 1.23 $
**
**  History file routines.
*/
#include <stdio.h>
#include <sys/types.h>
#include "configdata.h"
#include "clibrary.h"
#include "innd.h"
#include "dbz.h"
#include "timer.h"
#include "cache.h"

#define TMR_RETARG(a, b)	{ TMRstop((a)); return (b); }

#if	(HIS_CACHE_SIZE > 0)

STATIC BOOL		HIScacheInitialized = FALSE;
unsigned long		HIScacheLookups;
unsigned long		HIScachePos;
unsigned long		HIScacheNeg;
unsigned long		HIScacheDBZPos;
unsigned long		HIScacheDBZNeg;
#endif	/* (HIS_CACHE_SIZE > 0) */

STATIC char	HIShistpath[] = _PATH_HISTORY;
STATIC FILE	*HISwritefp;
STATIC int	HISreadfd;
STATIC int	HISdirty;
STATIC int	HISincore = INND_DBZINCORE;


/*
**  Set up the history files.
*/
void
HISsetup()
{
    if (HISwritefp == NULL) {
	/* Open the history file for appending formatted I/O. */
	if ((HISwritefp = fopen(HIShistpath, "a")) == NULL) {
	    syslog(L_FATAL, "%s cant fopen %s %m", LogName, HIShistpath);
	    exit(1);
	}
	CloseOnExec((int)fileno(HISwritefp), TRUE);

	/* Open the history file for reading. */
	if ((HISreadfd = open(HIShistpath, O_RDONLY)) < 0) {
	    syslog(L_FATAL, "%s cant open %s %m", LogName, HIShistpath);
	    exit(1);
	}
	CloseOnExec(HISreadfd, TRUE);

	/* Open the DBZ file. */
	(void)dbzincore(HISincore);
	(void)dbzwritethrough(1);
	if (dbminit(HIShistpath) < 0) {
	    syslog(L_FATAL, "%s cant dbminit %s %m", HIShistpath, LogName);
	    exit(1);
	}
    }
/*
 * Initialize history caches
 */
#if	(HIS_CACHE_SIZE > 0)

    /* Cache can be same through the life of innd. */
    if (HIScacheInitialized == FALSE) {
	HIScacheInitialized = TRUE;
	CCHinit(HIShistpath);
    }
    /* Protect overflow by resetting counts at every history setup */
    HIScacheLookups = 0;
    HIScachePos = 0;
    HIScacheNeg = 0;
    HIScacheDBZPos = 0;
    HIScacheDBZNeg = 0;
#endif	/* (HIS_CACHE_SIZE > 0) */
}


/*
**  Synchronize the in-core history file (flush it).
*/
void
HISsync()
{
    TMRstart(TMR_HISSYNC);
    if (HISdirty) {
	if (dbzsync()) {
	    syslog(L_FATAL, "%s cant dbzsync %m", LogName);
	    exit(1);
	}
	HISdirty = 0;
    }
    TMRstop(TMR_HISSYNC);
}

char *
HISgetCacheStat()
{
#if	(HIS_CACHE_SIZE > 0)
    static char buf[4096];

    sprintf(buf, 
	"%lu looks %lu pos %lu neg %lu dpos %lu dneg",
	HIScacheLookups, HIScachePos, HIScacheNeg,
	HIScacheDBZPos, HIScacheDBZNeg);
    return buf;
#else
    return NULL;
#endif
}


/*
**  Close the history files.
*/
void
HISclose()
{
    if (HISwritefp != NULL) {
	/* Since dbmclose calls dbzsync we could replace this line with
	 * "HISdirty = 0;".  Oh well, it keeps the abstraction clean. */
	HISsync();
	if (dbmclose() < 0)
	    syslog(L_ERROR, "%s cant dbmclose %m", LogName);
	if (fclose(HISwritefp) == EOF)
	    syslog(L_ERROR, "%s cant fclose history %m", LogName);
	HISwritefp = NULL;
	if (close(HISreadfd) < 0)
	    syslog(L_ERROR, "%s cant close history %m", LogName);
	HISreadfd = -1;
    }
#if	(HIS_CACHE_SIZE > 0)
    CCHsync(HIShistpath);
    syslog(L_NOTICE, "SERVER hiscache: %s\n", HISgetCacheStat() );
#endif	/* (HIS_CACHE_SIZE > 0) */
}

/*
**  File in the DBZ datum for a Message-ID, making sure not to copy any
**  illegal characters.
*/
STATIC void
HISsetkey(p, keyp)
    register char	*p;
    datum		*keyp;
{
    static BUFFER	MessageID;
    register char	*dest;
    register int	i;

    /* Get space to hold the ID. */
    i = strlen(p);
    if (MessageID.Data == NULL) {
	MessageID.Data = NEW(char, i + 1);
	MessageID.Size = i;
    }
    else if (MessageID.Size < i) {
	RENEW(MessageID.Data, char, i + 1);
	MessageID.Size = i;
    }

    for (keyp->dptr = dest = MessageID.Data; *p; p++)
	if (*p == HIS_FIELDSEP || *p == '\n')
	    *dest++ = HIS_BADCHAR;
	else
	    *dest++ = *p;
    *dest = '\0';

    keyp->dsize = dest - MessageID.Data + 1;
}


/*
**  Get the list of files under which a Message-ID is stored.
*/
char *
HISfilesfor(char *MessageID)
{
    static BUFFER	Files;
    char		*dest;
    datum		key;
    datum		val;
    long		offset;
    register char	*p;
    register int	i;

    /* Get the seek value into the history file. */
    TMRstart(TMR_HISGREP);
    HISsetkey(MessageID, &key);
    val = dbzfetch(key);

    if (val.dptr == NULL || val.dsize != sizeof offset)
	TMR_RETARG(TMR_HISGREP, NULL);

    /* Get space. */
    if (Files.Data == NULL) {
	Files.Size = BUFSIZ;
	Files.Data = NEW(char, Files.Size);
    }

    /* Copy the value to an aligned spot. */
    for (p = val.dptr, dest = (char *)&offset, i = sizeof offset; --i >= 0; )
	*dest++ = *p++;
    if (lseek(HISreadfd, offset, SEEK_SET) == -1)
	TMR_RETARG(TMR_HISGREP, NULL);

    /* Read the text until \n or EOF. */
    for (Files.Used = 0; ; ) {
	i = read(HISreadfd,
		&Files.Data[Files.Used], Files.Size - Files.Used - 1);
	if (i <= 0)
	    TMR_RETARG(TMR_HISGREP, NULL);
	Files.Used += i;
	Files.Data[Files.Used] = '\0';
	if ((p = strchr(Files.Data, '\n')) != NULL) {
	    *p = '\0';
	    break;
	}

	/* If we have half our buffer left, get more space. */
	if (Files.Size - Files.Used < Files.Size / 2) {
	    Files.Size += BUFSIZ;
	    RENEW(Files.Data, char, Files.Size);
	}
    }

    /* Move past the first two fields -- Message-ID and date info. */
    if ((p = strchr(Files.Data, HIS_FIELDSEP)) == NULL)
	TMR_RETARG(TMR_HISGREP, NULL);
    if ((p = strchr(p + 1, HIS_FIELDSEP)) == NULL)
	TMR_RETARG(TMR_HISGREP, NULL);

    /* Translate newsgroup separators to slashes, return the fieldstart. */
    for (dest = ++p; *p; p++)
	if (*p == '.')
	    *p = '/';
    TMR_RETARG(TMR_HISGREP, dest);
}

/*
**  Have we already seen an article?
*/
BOOL
HIShavearticle(char *MessageID)
{
    datum	key;
    datum	val;
    cachelook_t	status;

    if (MessageID == NULL || *MessageID == '\0')
	return TRUE;

    TMRstart(TMR_HISHAVE);
#if	(HIS_CACHE_SIZE > 0)
    status = CCHhavearticle(MessageID);
    HIScacheLookups++;
    if (status == CCH_HAVEART) {
	HIScachePos++;
	TMR_RETARG (TMR_HISHAVE, TRUE);
    }
    if (status == CCH_HAVEID) {
	HIScacheNeg++;
	TMR_RETARG (TMR_HISHAVE, FALSE);
    }
#endif

    HISsetkey(MessageID, &key);
    val = dbzfetch(key);

#if	(HIS_CACHE_SIZE > 0)
    if (val.dptr) {
	HIScacheDBZPos++;
	CCHstore((CSTRING)MessageID, Now.time, TRUE);
    }
    else {
	HIScacheDBZNeg++;
	CCHstore(MessageID, Now.time, FALSE);
    }
#endif
    TMR_RETARG(TMR_HISHAVE, (val.dptr != NULL));
}

/*
**  Turn a history filename entry from slashes to dots.  It's a pity
**  we have to do this.
*/
STATIC void
HISslashify(p)
    register char	*p;
{
    register char	*last;

    for (last = NULL; *p; p++) {
	if (*p == '/') {
	    *p = '.';
	    last = p;
	}
	else if (*p == ' ' && last != NULL)
	    *last = '/';
    }
    if (last)
	*last = '/';
}


/*
**  Write a history entry.
*/
BOOL
HISwrite(ARTDATA *Data, char *paths)
{
    static char		NOPATHS[] = "";
    long		offset;
    datum		key;
    datum		val;
    int			i;

    TMRstart(TMR_HISWRITE);
    HISsetkey(Data->MessageID, &key);
    if (paths != NULL && paths[0] != '\0')
	HISslashify(paths);
    else
	paths = NOPATHS;

    offset = ftell(HISwritefp);
    if (Data->Expires > 0)
	i = fprintf(HISwritefp, "%s%c%lu%c%lu%c%lu%c%s\n",
                    key.dptr, HIS_FIELDSEP,
                    (unsigned long)Data->Arrived, HIS_SUBFIELDSEP,
                    (unsigned long)Data->Expires,
		    HIS_SUBFIELDSEP, (unsigned long)Data->Posted, HIS_FIELDSEP,
                    paths);
    else
	i = fprintf(HISwritefp, "%s%c%lu%c%s%c%lu%c%s\n",
                    key.dptr, HIS_FIELDSEP,
                    (unsigned long)Data->Arrived,
                    HIS_SUBFIELDSEP, HIS_NOEXP,
		    HIS_SUBFIELDSEP, (unsigned long)Data->Posted, HIS_FIELDSEP,
                    paths);
    if (i == EOF || fflush(HISwritefp) == EOF) {
	/* The history line is now an orphan... */
	i = errno;
	syslog(L_ERROR, "%s cant write history %m", LogName);
	IOError("history", i);
	TMR_RETARG(TMR_HISWRITE, FALSE);
    }

    /* Set up the database values and write them. */
    val.dptr = (char *)&offset;
    val.dsize = sizeof offset;
    if (dbzstore(key, val) < 0) {
	i = errno;
	syslog(L_ERROR, "%s cant dbzstore %m", LogName);
	IOError("history database", i);
	TMR_RETARG(TMR_HISWRITE, FALSE);
    }

#if	(HIS_CACHE_SIZE > 0)
    CCHstore((CSTRING)Data->MessageID, Now.time, TRUE);
#endif	/* (HIS_CACHE_SIZE > 0) */

    if (++HISdirty >= ICD_SYNC_COUNT)
	HISsync();
    TMR_RETARG(TMR_HISWRITE, TRUE);
}
